#include <iostream>
#include <vcg/complex/complex.h>

#include <wrap/io_trimesh/import.h>
#include <wrap/io_trimesh/export.h>

#include <vcg/space/index/kdtree/kdtree.h>
#include <vcg/space/index/grid_static_ptr.h>
#include <vcg/space/index/spatial_hashing.h>
#include <vcg/space/index/octree.h>

int num_test = 1000;
int kNearest = 256;
float queryDist = 0.0037;
float bboxratio = 1000.0f;

using namespace vcg;


class CVertex;
class CFace;
class CEdge;

class CUsedTypes	: public vcg::UsedTypes < vcg::Use< CVertex >::AsVertexType, vcg::Use< CFace >::AsFaceType>{};
class CVertex		: public vcg::Vertex < CUsedTypes, vcg::vertex::Coord3f, vcg::vertex::Normal3f, vcg::vertex::Radiusf, vcg::vertex::BitFlags, vcg::vertex::Qualityf, vcg::vertex::Color4b>{};
class CFace			: public vcg::Face < CUsedTypes, vcg::face::VertexRef>{};

class CMesh			: public vcg::tri::TriMesh < std::vector< CVertex >, std::vector< CFace > > {};

int elapsed(int t)
{
    return ((clock()-t)*1000.0)/CLOCKS_PER_SEC;
}


void testKDTree(CMesh& mesh, std::vector<unsigned int>& test_indeces, std::vector<vcg::Point3f>& randomSamples)
{
    std::cout << "==================================================="<< std::endl;
    std::cout << "KDTree" << std::endl;
    int t0=clock();
    // Construction of the kdTree
    vcg::ConstDataWrapper<CMesh::VertexType::CoordType> wrapperVcg(&mesh.vert[0].P(), mesh.vert.size(), size_t(mesh.vert[1].P().V()) - size_t(mesh.vert[0].P().V()));
    vcg::KdTree<CMesh::ScalarType> kdTreeVcg(wrapperVcg);
    std::cout << "Build: " << elapsed(t0) <<  " ms" << std::endl;
    int nn=1;
    // Computation of the average distance between points the mesh
    float mAveragePointSpacing = 0;
    t0=clock();
    for (size_t i = 0; i < mesh.vert.size(); i++)
    {
        vcg::KdTree<CMesh::ScalarType>::PriorityQueue queue;
        kdTreeVcg.doQueryK(mesh.vert[i].cP(), 16, queue);
        float newRadius = 2.0f * sqrt(queue.getWeight(0)/ queue.getNofElements());
        mesh.vert[i].R() -= newRadius;
        mAveragePointSpacing += newRadius;
    }
    std::cout << "Num trhread " << nn << std::endl;
    mAveragePointSpacing /= mesh.vert.size();
    std::cout << "Average point radius (OpenMP with" << nn << " threads) " << mAveragePointSpacing << std::endl;
    std::cout << "Time (OpenMP): " << elapsed(t0) << " ms" << std::endl;
    
    queryDist = mAveragePointSpacing * 150;
    
    // Test with the radius search
    std::cout << "Radius search (" << num_test << " tests)"<< std::endl;
    float avgTime = 0.0f;
    for (int ii = 0; ii < num_test; ii++)
    {
        int t0=clock();
        std::vector<unsigned int> indeces;
        std::vector<float> dists;
        kdTreeVcg.doQueryDist(mesh.vert[test_indeces[ii]].cP(), queryDist, indeces, dists);
        avgTime += elapsed(t0);
    }
    std::cout << "Time (radius = " << queryDist << "): " << avgTime << " ms (mean " << avgTime / num_test << "ms)"  << std::endl;
    
    // Test with the k-nearest search
    std::cout << "k-Nearest search (" << num_test*10 << " tests)"<< std::endl;
    avgTime = 0.0f;
    for (int ii = 0; ii < num_test * 10; ii++)
    {
        int t0=clock();
        vcg::KdTree<CMesh::ScalarType>::PriorityQueue queue;
        kdTreeVcg.doQueryK(mesh.vert[test_indeces[ii]].cP(), kNearest, queue);
        avgTime += elapsed(t0);
    }
    std::cout << "Time (k = " << kNearest << "): " << avgTime << " ms (mean " << avgTime / (num_test * 10) << "ms)"  << std::endl;
    
    // Test with the closest search
    std::cout << "Closest search (" << num_test*10 << " tests)"<< std::endl;
    avgTime = 0.0f;
    for (int ii = 0; ii < num_test * 10; ii++)
    {
        int t0=clock();
        unsigned int index;
        float minDist;
        kdTreeVcg.doQueryClosest(randomSamples[ii], index, minDist);
        avgTime += elapsed(t0);
    }
    std::cout << "Time : " << avgTime << " ms (mean " << avgTime / (num_test * 10) << "ms)"  << std::endl << std::endl;
}


void testUniformGrid(CMesh& mesh, std::vector<unsigned int>& test_indeces, std::vector<vcg::Point3f>& randomSamples)
{
    std::cout << "==================================================="<< std::endl;
    std::cout << "Uniform Grid" << std::endl;
    int t0=clock();
    
    // Construction of the uniform grid
    typedef vcg::GridStaticPtr<CMesh::VertexType, CMesh::VertexType::ScalarType> MeshGrid;
    MeshGrid uniformGrid;
    uniformGrid.Set(mesh.vert.begin(), mesh.vert.end());
    std::cout << "Build: " << elapsed(t0) <<  " ms" << std::endl;
    
    // Test with the radius search
    std::cout << "Radius search (" << num_test << " tests)"<< std::endl;
    float  avgTime = 0.0f;
    for (int ii = 0; ii < num_test; ii++)
    {
        t0=clock();
        std::vector<CMesh::VertexPointer> vertexPtr;
        std::vector<CMesh::VertexType::CoordType> points;
        std::vector<float> dists;
        vcg::tri::GetInSphereVertex(mesh, uniformGrid, mesh.vert[test_indeces[ii]].cP(), queryDist, vertexPtr, dists, points);
        avgTime += elapsed(t0);
    }
    std::cout << "Time (radius = " << queryDist << "): " << avgTime << " ms (mean " << avgTime / num_test << "ms)"  << std::endl;
    
    // Test with the k-nearest search
    std::cout << "k-Nearest search (" << num_test*10 << " tests)"<< std::endl;
    avgTime = 0.0f;
    for (int ii = 0; ii < num_test * 10; ii++)
    {
        t0=clock();
        std::vector<CMesh::VertexPointer> vertexPtr;
        std::vector<CMesh::VertexType::CoordType> points;
        std::vector<float> dists;
        vcg::tri::GetKClosestVertex(mesh, uniformGrid, kNearest, mesh.vert[test_indeces[ii]].cP(), mesh.bbox.Diag(), vertexPtr, dists, points);
        avgTime += elapsed(t0);
    }
    std::cout << "Time (k = " << kNearest << "): " << avgTime << " ms (mean " << avgTime / (num_test * 10) << "ms)"  << std::endl;
    
    // Test with the Closest search
    std::cout << "Closest search (" << num_test*10 << " tests)"<< std::endl;
    avgTime = 0.0f;
    for (int ii = 0; ii < num_test * 10; ii++)
    {
        t0=clock();
        float minDist;
        vcg::tri::GetClosestVertex(mesh, uniformGrid, randomSamples[ii], mesh.bbox.Diag(), minDist);
        avgTime += elapsed(t0);
    }
    std::cout << "Time : " << avgTime << " ms (mean " << avgTime / (num_test * 10) << "ms)"  << std::endl << std::endl;
}



void testSpatialHashing(CMesh& mesh, std::vector<unsigned int>& test_indeces, std::vector<vcg::Point3f>& randomSamples)
{
    std::cout << "==================================================="<< std::endl;
    std::cout << "Spatial Hashing" << std::endl;
    int t0=clock();
    
    // Construction of the uniform grid
    typedef vcg::SpatialHashTable<CMesh::VertexType, CMesh::VertexType::ScalarType> MeshGrid;
    MeshGrid uniformGrid;
    uniformGrid.Set(mesh.vert.begin(), mesh.vert.end());
    std::cout << "Build: " << elapsed(t0) <<  " ms" << std::endl;
    
    // Test with the radius search
    std::cout << "Radius search (" << num_test << " tests)"<< std::endl;
    float  avgTime = 0.0f;
    for (int ii = 0; ii < num_test; ii++)
    {
        t0=clock();
        std::vector<CMesh::VertexPointer> vertexPtr;
        std::vector<CMesh::VertexType::CoordType> points;
        std::vector<float> dists;
        vcg::tri::GetInSphereVertex(mesh, uniformGrid, mesh.vert[test_indeces[ii]].cP(), queryDist, vertexPtr, dists, points);
        avgTime += elapsed(t0);
    }
    std::cout << "Time (radius = " << queryDist << "): " << avgTime << " ms (mean " << avgTime / num_test << "ms)"  << std::endl;
    
    // Test with the k-nearest search
    std::cout << "k-Nearest search (" << num_test*10 << " tests)"<< std::endl;
    avgTime = 0.0f;
    for (int ii = 0; ii < num_test * 10; ii++)
    {
        t0=clock();
        std::vector<CMesh::VertexPointer> vertexPtr;
        std::vector<CMesh::VertexType::CoordType> points;
        std::vector<float> dists;
        vcg::tri::GetKClosestVertex(mesh, uniformGrid, kNearest, mesh.vert[test_indeces[ii]].cP(), mesh.bbox.Diag(), vertexPtr, dists, points);
        avgTime += elapsed(t0);
    }
    std::cout << "Time (k = " << kNearest << "): " << avgTime << " ms (mean " << avgTime / (num_test * 10) << "ms)"  << std::endl;
    
    // Test with the Closest search
    std::cout << "Closest search (" << num_test*10 << " tests)"<< std::endl;
    avgTime = 0.0f;
    for (int ii = 0; ii < num_test * 10; ii++)
    {
        t0=clock();
        float minDist;
        vcg::tri::GetClosestVertex(mesh, uniformGrid, randomSamples[ii], mesh.bbox.Diag(), minDist);
        avgTime += elapsed(t0);
    }
    std::cout << "Time : " << avgTime << " ms (mean " << avgTime / (num_test * 10) << "ms)"  << std::endl << std::endl;
}




int main( int argc, char * argv[] )
{
    if (argc < 2) {
        std::cout << "Invalid arguments" << std::endl;
        exit(-1);
    }
    CMesh mesh;
    if (vcg::tri::io::Importer<CMesh>::Open(mesh, argv[1])  != 0)
        std::cout << "Invalid filename" << std::endl;
    
    std::cout << "Mesh BBox diagonal: " << mesh.bbox.Diag() << std::endl;
    std::cout << "Max point random offset: " << mesh.bbox.Diag() / 1000.0f << std::endl << std::endl;
    
    vcg::math::MarsenneTwisterRNG randGen;
    randGen.initialize(0);
    std::vector<vcg::Point3f> randomSphereSamples;
    CMesh samplePointsMesh;
    for (int i = 0; i < num_test * 10; i++)
    {
        randomSphereSamples.push_back(vcg::math::GeneratePointOnUnitSphereUniform<float>(randGen) * randGen.generate01() * mesh.bbox.Diag()/2.0f);
        vcg::tri::Allocator<CMesh>::AddVertex(samplePointsMesh,randomSphereSamples.back());
    }
    tri::io::ExporterPLY<CMesh>::Save(samplePointsMesh, "pointsSphere.ply");
    
    // build a set of points on the surface of the mesh with a not uniform distribution
    // the probability of a point to be on a triangle should be proportional to the area of the triangle
    std::vector<vcg::Point3f> randomSurfaceSamples;
    CMesh surfacePointsMesh;
    for (int i = 0; i < num_test * 10; i++)
    {
        int triIndex = randGen.generate(mesh.FN());
        CMesh::FaceType &f = mesh.face[triIndex];
        float a1 = randGen.generate01();
        float a2 = randGen.generate01();
        if(a1+a2>1.0)
        {
            a1=1.0f-a1;
            a2=1.0f-a2;
        }
        Point3f randPoint = f.P(0)+(f.P(1) - f.P(0))*a1 + (f.P(2)-f.P(0))*a2;
        randomSurfaceSamples.push_back(randPoint);
        vcg::tri::Allocator<CMesh>::AddVertex(surfacePointsMesh,randomSurfaceSamples.back());
    }
    tri::io::ExporterPLY<CMesh>::Save(surfacePointsMesh, "pointsSurface.ply");
    
    std::vector<unsigned int> test_indeces;
    for (int i = 0; i < num_test * 10; i++)
    {
        int index = randGen.generate01() * (mesh.vert.size() - 1);
        test_indeces.push_back(index);
        randomSphereSamples[i] += mesh.vert[i].P();
    }
    
    testKDTree(mesh, test_indeces, randomSurfaceSamples);
    testUniformGrid(mesh, test_indeces, randomSurfaceSamples);
    testSpatialHashing(mesh, test_indeces, randomSurfaceSamples);
}
